import mlx.core as mx
from mlx_lm.tokenizer_utils import TokenizerWrapper
from outlines.models.transformers import TransformerTokenizer
from outlines.processors import JSONLogitsProcessor

from ..schema import ResponseFormat


class OutlinesLogitsProcessor:
    processed_token_count: int = 0

    def __init__(self, tokenizer: TokenizerWrapper, response_format: ResponseFormat):
        json_schema = response_format.json_schema.schema_def
        self.logits_processor = JSONLogitsProcessor(
            json_schema,
            TransformerTokenizer(tokenizer._tokenizer),
            tensor_library_name="mlx",
        )

    def __call__(self, tokens: mx.array, logits: mx.array) -> mx.array:
        # `tokens` is the full sequence of tokens (prompt + generated).
        # The Outlines processor is stateful and needs the full sequence.

        # Outlines processor expects 1D logits
        logits_1d = logits.reshape(-1)

        # Ensure logits are float32 for the processor
        # The result from outlines is a numpy array, convert it back to mx.array
        processed_logits = self.logits_processor(tokens, logits_1d.astype(mx.float32))

        # Reshape to the original logits shape.
        return mx.array(processed_logits).reshape(logits.shape)



"""
Prompt Cache Management Module

This module provides functionality for managing and optimizing model prompt caching,
to improve performance in multi-turn conversations.
"""

from dataclasses import dataclass, field
from typing import Any, List

from mlx_lm.models.cache import (
    can_trim_prompt_cache,
    make_prompt_cache,
    trim_prompt_cache,
)

from mlx_omni_server.chat.mlx.model_types import MlxModelCache

from ...utils.logger import logger


def common_prefix_len(list1, list2):
    """
    Calculates the length of the common prefix of two lists.

    Args:
        list1: The first list of strings.
        list2: The second list of strings.

    Returns:
        The length of the common prefix. Returns 0 if lists are empty
        or do not match at the first element.
    """
    # Determine the maximum possible length of the common prefix
    min_len = min(len(list1), len(list2))

    # Iterate up to the length of the shorter list
    for i in range(min_len):
        if list1[i] != list2[i]:
            # Mismatch found, the common prefix length is the current index
            return i

    # No mismatch found within the bounds of the shorter list,
    # so the common prefix length is the length of the shorter list.
    return min_len


@dataclass
class PromptCache:
    """
    Prompt cache class for storing and managing model prompt caches

    Attributes:
        tokens: Cached token sequence
        cache: Model's KV cache state, a list matching the number of model layers
        model_key: Model identifier to ensure cache matches the model
    """

    tokens: List[int] = field(default_factory=list)
    cache: List[Any] = field(default_factory=list)
    model_key: str = ""

    def extend_completion_cache(self, completion_tokens):
        self.tokens.extend(completion_tokens)

    def reset_prompt_cache(self, model_cache: MlxModelCache, prompt):
        logger.debug("*** Resetting cache. ***")
        self.model_key = model_cache.model_id.name
        self.cache = make_prompt_cache(model_cache.model)

        if model_cache.draft_model is not None:
            self.cache += make_prompt_cache(model_cache.draft_model)

        self.tokens = list(prompt)  # Cache the new prompt fully

    def get_prompt_cache(self, model_cache, prompt):
        """
        Determines the portion of the prompt that needs processing by comparing
        it to the cached prompt and attempting to reuse the common prefix.

        This function updates the internal prompt cache state (tokens and model cache)
        based on the comparison. If a common prefix exists, it attempts to trim
        the model cache (if supported) to match the common prefix length, avoiding
        recomputation.

        Args:
            prompt (List[int]): The tokenized new prompt.

        Returns:
            List[int]: The suffix of the prompt that actually needs to be processed
                       by the model. This will be the full prompt if the cache is
                       reset or cannot be effectively used.
        """
        cache_len = len(self.tokens)
        prompt_len = len(prompt)
        com_prefix_len = common_prefix_len(self.tokens, prompt)
        prompt_cached_tokens = 0

        # Leave at least one token in the prompt
        com_prefix_len = min(com_prefix_len, len(prompt) - 1)

        # Condition 1: Model changed or no common prefix at all. Reset cache.
        if self.model_key != model_cache.model_id.name or com_prefix_len == 0:
            self.reset_prompt_cache(model_cache, prompt)

        # Condition 2: Common prefix exists and matches cache length. Process suffix.
        elif com_prefix_len == cache_len:
            logger.debug(
                f"*** Cache is prefix of prompt (cache_len: {cache_len}, prompt_len: {prompt_len}). Processing suffix. ***"
            )
            prompt = prompt[com_prefix_len:]
            self.tokens.extend(prompt)
            prompt_cached_tokens = com_prefix_len

        # Condition 3: Common prefix exists but is shorter than cache length. Attempt trim.
        elif com_prefix_len < cache_len:
            logger.debug(
                f"*** Common prefix ({com_prefix_len}) shorter than cache ({cache_len}). Attempting trim. ***"
            )

            if can_trim_prompt_cache(self.cache):
                num_to_trim = cache_len - com_prefix_len
                logger.debug(f"    Trimming {num_to_trim} tokens from cache.")
                trim_prompt_cache(self.cache, num_to_trim)
                self.tokens = self.tokens[:com_prefix_len]
                prompt = prompt[com_prefix_len:]
                self.tokens.extend(prompt)
                prompt_cached_tokens = com_prefix_len
            else:
                logger.debug("    Cache cannot be trimmed. Resetting cache.")
                self.reset_prompt_cache(model_cache, prompt)

        # This case should logically not be reached if com_prefix_len <= cache_len
        else:
            logger.error(
                f"Unexpected cache state: com_prefix_len ({com_prefix_len}) > cache_len ({cache_len}). Resetting cache."
            )
            self.reset_prompt_cache(model_cache, prompt)

        logger.debug(f"Returning {len(prompt)} tokens for processing.")
        return prompt, prompt_cached_tokens



from dataclasses import dataclass
from typing import Optional, Type

import mlx.nn as nn
from mlx_lm.tokenizer_utils import TokenizerWrapper
from mlx_lm.utils import get_model_path, load, load_config

from ...utils.logger import logger
from .tools.chat_tokenizer import ChatTokenizer
from .tools.hugging_face import HuggingFaceChatTokenizer
from .tools.llama3 import Llama3ChatTokenizer
from .tools.mistral import MistralChatTokenizer


def load_tools_handler(model_type: str, tokenizer: TokenizerWrapper) -> ChatTokenizer:
    """Factory function to load appropriate tools handler based on model ID."""
    handlers: dict[str, Type[ChatTokenizer]] = {
        # Llama models
        "llama": Llama3ChatTokenizer,
        "mistral": MistralChatTokenizer,
        "qwen2": HuggingFaceChatTokenizer,
    }

    # Get handler class based on model ID or use Llama handler as default
    handler_class = handlers.get(model_type, HuggingFaceChatTokenizer)
    return handler_class(tokenizer)


@dataclass(frozen=True)
class ModelId:
    """Entity class representing a unique model identifier.

    This class encapsulates all parameters that determine whether a model
    needs to be reloaded. It replaces the tuple-based approach with a more
    object-oriented design that's easier to extend and maintain.
    """

    name: str
    adapter_path: Optional[str] = None
    draft_model: Optional[str] = None

    def __str__(self) -> str:
        """Return a string representation of the model ID for debugging."""
        parts = [f"model_name={self.name}"]
        if self.adapter_path:
            parts.append(f"adapter_path={self.adapter_path}")
        if self.draft_model:
            parts.append(f"draft_model={self.draft_model}")
        return f"ModelId({', '.join(parts)})"


class MlxModelCache:
    """Model cache class to avoid reloading the same models.

    This class manages the cache of main models and draft models based on ModelId objects.
    """

    def __init__(self, model_id: ModelId):
        """Initialize the cache object.

        Args:
            model_id: Optional ModelId object for initialization
        """
        self.model_id: ModelId = model_id
        self.model: Optional[nn.Module] = None
        self.tokenizer: Optional[TokenizerWrapper] = None
        self.chat_tokenizer = None
        self.draft_model: Optional[nn.Module] = None
        self.draft_tokenizer: Optional[TokenizerWrapper] = None

        # If a model ID is provided, load the models directly
        if model_id:
            self._load_models()

    def _load_models(self):
        """Load the main model and draft model (if needed)."""
        # Load the main model
        self.model, self.tokenizer = load(
            self.model_id.name,
            tokenizer_config={"trust_remote_code": True},
            adapter_path=self.model_id.adapter_path,
        )
        logger.info(f"Loaded new model: {self.model_id.name}")

        # Load configuration and create chat tokenizer
        model_path = get_model_path(self.model_id.name)[0]
        config = load_config(model_path)
        self.chat_tokenizer = load_tools_handler(config["model_type"], self.tokenizer)

        # If needed, load the draft model
        if self.model_id.draft_model:
            self.draft_model, self.draft_tokenizer = load(
                self.model_id.draft_model,
                tokenizer_config={"trust_remote_code": True},
            )

            # Check if vocabulary sizes match
            if self.draft_tokenizer.vocab_size != self.tokenizer.vocab_size:
                logger.warn(
                    f"Draft model({self.model_id.draft_model}) tokenizer does not match model tokenizer."
                )

            logger.info(f"Loaded new draft model: {self.model_id.draft_model}")



from typing import Any, Dict, List, Optional

from pydantic import BaseModel, Field, model_serializer


class Model(BaseModel):
    """Model information as per OpenAI API specification"""

    id: str = Field(..., description="The model identifier")
    object: str = Field(default="model", description="The object type (always 'model')")
    created: int = Field(
        ..., description="Unix timestamp of when the model was created"
    )
    owned_by: str = Field(..., description="Organization that owns the model")
    details: Optional[Dict[str, Any]] = Field(
        default=None, description="Full model configuration (if details are requested)"
    )

    @model_serializer
    def serialize_model(self):
        """Custom serializer to exclude None details field"""
        data = {
            "id": self.id,
            "object": self.object,
            "created": self.created,
            "owned_by": self.owned_by,
        }
        if self.details is not None:
            data["details"] = self.details
        return data


class ModelList(BaseModel):
    """Response format for list of models"""

    object: str = Field(default="list", description="The object type (always 'list')")
    data: List[Model] = Field(..., description="List of model objects")


class ModelDeletion(BaseModel):
    """Response format for model deletion"""

    id: str = Field(..., description="The ID of the deleted model")
    object: str = Field(default="model", description="The object type (always 'model')")
    deleted: bool = Field(..., description="Whether the model was deleted")



import re
from abc import ABC, abstractmethod
from typing import Any, Dict, List, Optional

from mlx_lm.tokenizer_utils import TokenizerWrapper


class TokensDecoder(ABC):
    """Base class for tokens decoders."""

    @abstractmethod
    def decode(self, text: str) -> Optional[Dict[str, Any]]:
        """Parse tool calls from model output."""
        pass

    def stream_decode(self, text: str) -> Optional[Dict[str, Any]]:
        """Parse tool calls from model output."""
        return self.decode(text)


class ReasoningDecoder(TokensDecoder):
    """Base class for reasoning decoders."""

    thinking_tag: str = "think"
    enable_thinking: bool = False
    add_thinking_prefix: bool = False
    accumulated_text: str = ""

    def __init__(self, tokenizer: TokenizerWrapper):
        self.tokenizer = tokenizer
        self.accumulated_text = ""

    def set_thinking_prefix(self, add_thinking_prefix: bool) -> None:
        self.add_thinking_prefix = add_thinking_prefix
        self.accumulated_text = ""
        if add_thinking_prefix:
            self.accumulated_text = f"<{self.thinking_tag}>"

    def _parse_stream_response(self, text: str) -> Optional[Dict[str, Any]]:
        # Check if in thinking mode
        thinking_end_tag = f"</{self.thinking_tag}>"
        thinking_start_tag = f"<{self.thinking_tag}>"

        # Special case: text exactly equals the start tag
        if text == thinking_start_tag:
            self.accumulated_text += text
            return {"delta_content": None, "delta_reasoning": ""}

        # Special case: text exactly equals the end tag
        if text == thinking_end_tag:
            self.accumulated_text += text
            return {
                "delta_content": "",
                "delta_reasoning": None,
            }

        # Update accumulated text
        self.accumulated_text += text

        # Check if accumulated text already contains the end tag
        has_end_tag_before = thinking_end_tag in self.accumulated_text[: -len(text)]
        has_end_tag_now = thinking_end_tag in self.accumulated_text

        # If text starts with thinking tag and end tag hasn't been encountered yet
        if self.accumulated_text.startswith(thinking_start_tag) and not has_end_tag_now:
            # delta content as reasoning
            return {"delta_content": None, "delta_reasoning": text}
        # If current delta contains the end tag (from not having it to having it)
        elif not has_end_tag_before and has_end_tag_now:
            # Split the current delta
            parts = text.split(thinking_end_tag, 1)
            return {"delta_content": "", "delta_reasoning": None}
        # If end tag was already encountered before
        elif has_end_tag_before:
            # All content after the end tag is content
            return {"delta_content": text, "delta_reasoning": None}
        else:
            # Other cases, possibly thinking mode not enabled or other situations
            return {"delta_content": text, "delta_reasoning": None}

    def stream_decode(self, text: str) -> Optional[Dict[str, Any]]:
        """Parse tool calls from model output."""
        if not self.enable_thinking:
            return {"delta_content": text}

        return self._parse_stream_response(text)

    def _parse_response(self, response: str):
        tag = self.thinking_tag
        # First check for complete thinking tag pattern
        reasoning_regex = f"<{tag}>([\s\S]*?)</{tag}>"
        reasoning_match = re.search(reasoning_regex, response)

        if reasoning_match:
            # Extract thinking content
            reasoning_content = reasoning_match.group(1).strip()

            # Get final content by replacing thinking tag and its content
            content = re.sub(reasoning_regex, "", response, count=1).strip()

            return {
                "content": content,
                "reasoning": reasoning_content,
            }
        else:
            # Check if only end tag exists (missing start tag case)
            end_tag = f"</{tag}>"
            if end_tag in response:
                # Split response using end tag
                parts = response.split(end_tag, 1)
                if len(parts) == 2:
                    reasoning_content = parts[0].strip()
                    content = parts[1].strip()
                    return {
                        "content": content,
                        "reasoning": reasoning_content,
                    }

            # If no tags exist, the entire response is the content
            return {
                "content": response.strip(),
                "reasoning": None,
            }

    def decode(self, text: str) -> Optional[Dict[str, Any]]:
        """Parse thinking content from model output"""
        if self.enable_thinking:
            return self._parse_response(text)

        return {"content": text}



import time
import uuid
from typing import Any, Dict, Generator, List, Optional

import mlx.core as mx
from mlx_lm.generate import GenerationResponse, stream_generate
from mlx_lm.sample_utils import make_logits_processors, make_sampler
from mlx_lm.tokenizer_utils import TokenizerWrapper

from ...utils.logger import logger
from ..schema import (
    ChatCompletionChoice,
    ChatCompletionChunk,
    ChatCompletionChunkChoice,
    ChatCompletionRequest,
    ChatCompletionResponse,
    ChatCompletionUsage,
    ChatMessage,
    Role,
)
from ..text_models import BaseTextModel, GenerateResult, GenerationParams
from .model_types import MlxModelCache
from .outlines_logits_processor import OutlinesLogitsProcessor
from .prompt_cache import PromptCache
from .tools.reasoning_decoder import ReasoningDecoder


class MLXModel(BaseTextModel):
    """MLX Chat Model wrapper with internal parameter management"""

    def __init__(
        self,
        model_cache: MlxModelCache,
    ):
        """Initialize MLXModel with model cache object.

        Args:
            model_cache: MlxModelCache object containing models and tokenizers
        """
        self._model_cache = model_cache
        self._default_max_tokens = 2048
        self._chat_tokenizer = model_cache.chat_tokenizer
        self._prompt_cache = PromptCache()
        self._prompt_cache_tokens_count = 0
        self._reasoning_decoder = ReasoningDecoder(model_cache.tokenizer)

    def _get_generation_params(
        self, request: ChatCompletionRequest
    ) -> GenerationParams:
        params = request.get_extra_params()

        # All params declare in `make_sampler`
        sampler_params = {
            "top_k",
            "min_tokens_to_keep",
            "min_p",
            "xtc_probability",
            "xtc_threshold",
            "xtc_special_tokens",
        }
        # Knowned params using in model config
        model_params = {
            "adapter_path",
            "draft_model",
            # Additional config for `apply_chat_template`
            "chat_template_config",
        }
        # Quick template params, same param will be overrided by `chat_template_config`
        template_params = {
            # Qwen3
            "enable_thinking",
            "thinking_budget",
            # Claude
            "thinking",
            # Gemini
            "thinkingConfig",
            # Grok
            "reasoning_effort",
            # Others
            "reasoning",
        }

        sampler_kwargs = {}
        model_kwargs = {}
        generate_kwargs = {}
        template_kwargs = {}

        for key, value in params.items():
            if key in sampler_params:
                sampler_kwargs[key] = value
            elif key in model_params:
                model_kwargs[key] = value
            elif key in template_params:
                template_kwargs[key] = value
            else:
                generate_kwargs[key] = value

        return {
            "sampler_kwargs": sampler_kwargs,
            "model_kwargs": model_kwargs,
            "generate_kwargs": generate_kwargs,
            "template_kwargs": template_kwargs,
        }

    def _process_logprobs(
        self,
        tokenizer: TokenizerWrapper,
        response: GenerationResponse,
        top_k: Optional[int],
    ) -> Optional[Dict[str, Any]]:
        """Process logprobs information from generation response to match OpenAI format"""
        current_token = response.token
        current_logprobs = response.logprobs

        # Get current token info
        token_str = tokenizer.decode([current_token])
        token_logprob = mx.clip(
            current_logprobs[current_token], a_min=-100, a_max=None
        ).item()
        token_bytes = token_str.encode("utf-8")

        # Base token info
        token_info = {
            "token": token_str,
            "logprob": token_logprob,
            "bytes": list(token_bytes),
        }

        # Process top logprobs
        top_logprobs = []
        if top_k is not None:
            # Get indices of top_k tokens
            top_indices = mx.argpartition(-current_logprobs, kth=top_k - 1)[:top_k]
            top_probs = mx.clip(current_logprobs[top_indices], a_min=-100, a_max=None)

            # Create detailed token information for each top token
            for idx, logprob in zip(top_indices.tolist(), top_probs.tolist()):
                token = tokenizer.decode([idx])
                token_bytes = token.encode("utf-8")
                top_logprobs.append(
                    {"token": token, "logprob": logprob, "bytes": list(token_bytes)}
                )

        return {**token_info, "top_logprobs": top_logprobs}

    def _prepare_generation(
        self,
        request: ChatCompletionRequest,
    ) -> tuple[Any, dict[str, Any]]:
        """Prepare all necessary components for generation.

        This function handles parameter processing, tokenizer setup, prompt encoding,
        sampler creation, and other preparation work for text generation.

        Args:
            request: The chat completion request containing generation parameters

        Returns:
            A tuple containing tokenizer, processed prompt, and generation kwargs
        """
        # Process parameters from request
        params = self._get_generation_params(request)

        model_kwargs = params.get("model_kwargs", {})
        logger.debug(f"Model kwargs: {model_kwargs}")

        template_kwargs = params.get("template_kwargs") | model_kwargs.get(
            "chat_template_config", {}
        )
        logger.debug(f"Chat Template kwargs: {template_kwargs}")

        # Prepare generation kwargs
        generate_kwargs = params.get("generate_kwargs", {})

        # Prepare sampler parameters
        sampler_kwargs = {
            "temp": (1.0 if request.temperature is None else request.temperature),
            "top_p": (1.0 if request.top_p is None else request.top_p),
            "min_p": 0.0,
            "min_tokens_to_keep": 1,
            "top_k": -1,
        } | params.get("sampler_kwargs", {})

        logger.debug(f"Sampler kwargs: {sampler_kwargs}")

        # Create sampler and add to generate_kwargs
        generate_kwargs["sampler"] = make_sampler(**sampler_kwargs)

        # Encode prompt with chat template
        prompt = self._chat_tokenizer.encode(
            messages=request.messages,
            tools=request.tools,
            **template_kwargs,
        )
        logger.debug(f"Encoded prompt:\n{prompt}")

        enable_thinking = template_kwargs.get("enable_thinking", True)
        self._reasoning_decoder.enable_thinking = enable_thinking
        if enable_thinking:
            self._reasoning_decoder.set_thinking_prefix(True)
            if prompt.endswith(f"<{self._reasoning_decoder.thinking_tag}>"):
                self._reasoning_decoder.set_thinking_prefix(True)
            else:
                self._reasoning_decoder.set_thinking_prefix(False)

        # Get tokenizer
        tokenizer = self._chat_tokenizer.tokenizer

        # Process prompt cache
        tokenized_prompt = tokenizer.encode(prompt)
        processed_prompt, cached_count = self._prompt_cache.get_prompt_cache(
            self._model_cache, tokenized_prompt
        )
        generate_kwargs["prompt_cache"] = self._prompt_cache.cache
        self._prompt_cache_tokens_count = cached_count
        logger.debug(
            f"Using {self._prompt_cache_tokens_count} cached tokens out of {len(tokenized_prompt)} total tokens"
        )

        # Setup logits processors
        if request.response_format and request.response_format.json_schema:
            generate_kwargs["logits_processors"] = [
                OutlinesLogitsProcessor(
                    self._chat_tokenizer.tokenizer, request.response_format
                )
            ]
        elif request.presence_penalty:
            generate_kwargs["logits_processors"] = make_logits_processors(
                repetition_penalty=request.presence_penalty
            )

        # Calculate max tokens for completion
        generate_kwargs["max_tokens"] = (
            request.max_completion_tokens
            or request.max_tokens
            or self._default_max_tokens
        )

        return processed_prompt, generate_kwargs

    def _stream_generate(
        self,
        request: ChatCompletionRequest,
    ) -> Generator[GenerateResult, None, None]:
        try:
            # Get tokenizer
            tokenizer = self._chat_tokenizer.tokenizer

            # Prepare all generation components
            processed_prompt, generate_kwargs = self._prepare_generation(request)

            generated_tokens = []
            for response in stream_generate(
                model=self._model_cache.model,
                tokenizer=tokenizer,
                prompt=processed_prompt,
                draft_model=self._model_cache.draft_model,
                **generate_kwargs,
            ):
                if response.finish_reason is not None:
                    break

                generated_tokens.append(response.token)

                logprobs = None
                if request.logprobs:
                    logprobs = self._process_logprobs(
                        tokenizer, response, request.top_logprobs
                    )

                yield GenerateResult(
                    text=response.text,
                    token=response.token,
                    finish_reason=response.finish_reason,
                    prompt_tokens=response.prompt_tokens,
                    generation_tokens=response.generation_tokens,
                    logprobs=logprobs,
                )

            if response is not None:
                logger.debug(
                    f"    prompt tokens: {response.prompt_tokens}, tps: {response.prompt_tps}"
                )
                logger.debug(
                    f"generation tokens: {response.generation_tokens}, tps: {response.generation_tps}"
                )

            self._prompt_cache.extend_completion_cache(generated_tokens)
        except Exception as e:
            logger.error(f"Error during stream generation: {str(e)}", exc_info=True)
            raise

    def generate(
        self,
        request: ChatCompletionRequest,
    ) -> ChatCompletionResponse:
        try:
            completion = ""
            logprobs_result_list = []
            generated_tokens = []
            finish_reason = "stop"
            result = None

            for result in self._stream_generate(request=request):
                generated_tokens.append(result.token)
                completion += result.text

                if request.logprobs:
                    logprobs_result_list.append(result.logprobs)

                if result.finish_reason:
                    finish_reason = result.finish_reason

            if result is None:
                raise RuntimeError("No tokens generated")

            logger.debug(f"Model Response:\n{completion}")
            reasoning: str | None = None  # avoid UnboundLocalError
            enable_thinking = self._reasoning_decoder.enable_thinking
            if enable_thinking:
                reasoning_result = self._reasoning_decoder.decode(completion)
                if reasoning_result:
                    logger.debug(f"Reasoning result:\n{reasoning_result}")
                    completion = reasoning_result.get("content")
                    reasoning = reasoning_result.get("reasoning") or None

            if request.tools:
                message = self._chat_tokenizer.decode(completion)
            else:
                message = ChatMessage(
                    role=Role.ASSISTANT,
                    content=completion,
                    reasoning=reasoning,
                )

            cached_tokens = self._prompt_cache_tokens_count
            logger.debug(f"Generate response with {cached_tokens} cached tokens")

            prompt_tokens_details = None
            if cached_tokens > 0:
                from ..schema import PromptTokensDetails

                prompt_tokens_details = PromptTokensDetails(cached_tokens=cached_tokens)

            return ChatCompletionResponse(
                id=f"chatcmpl-{uuid.uuid4().hex[:10]}",
                created=int(time.time()),
                model=request.model,
                choices=[
                    ChatCompletionChoice(
                        index=0,
                        message=message,
                        finish_reason=(
                            "tool_calls" if message.tool_calls else finish_reason
                        ),
                        logprobs=(
                            {"content": logprobs_result_list}
                            if logprobs_result_list
                            else None
                        ),
                    )
                ],
                usage=ChatCompletionUsage(
                    prompt_tokens=result.prompt_tokens + cached_tokens,
                    completion_tokens=result.generation_tokens,
                    total_tokens=result.prompt_tokens
                    + result.generation_tokens
                    + cached_tokens,
                    prompt_tokens_details=prompt_tokens_details,
                ),
            )
        except Exception as e:
            logger.error(f"Failed to generate completion: {str(e)}", exc_info=True)
            raise RuntimeError(f"Failed to generate completion: {str(e)}")

    def stream_generate(
        self,
        request: ChatCompletionRequest,
    ) -> Generator[ChatCompletionChunk, None, None]:
        try:
            chat_id = f"chatcmpl-{uuid.uuid4().hex[:10]}"

            completion = ""
            for result in self._stream_generate(request=request):
                created = int(time.time())
                completion += result.text

                message = None
                enable_thinking = self._reasoning_decoder.enable_thinking
                if enable_thinking:
                    reasoning_result = self._reasoning_decoder.stream_decode(
                        result.text
                    )
                    if reasoning_result:
                        logger.debug(f"Reasoning result:\n{reasoning_result}")
                        delta_content = reasoning_result.get("delta_content")
                        delta_reasoning = (
                            reasoning_result.get("delta_reasoning") or None
                        )
                        if delta_content:
                            message = ChatMessage(
                                role=Role.ASSISTANT,
                                content=delta_content or result.text,
                            )
                        elif delta_reasoning:
                            message = ChatMessage(
                                role=Role.ASSISTANT, reasoning=delta_reasoning
                            )

                if message is None:
                    message = ChatMessage(role=Role.ASSISTANT, content=result.text)

                yield ChatCompletionChunk(
                    id=chat_id,
                    created=created,
                    model=request.model,
                    choices=[
                        ChatCompletionChunkChoice(
                            index=0,
                            delta=message,
                            finish_reason=result.finish_reason,
                            logprobs=result.logprobs,
                        )
                    ],
                )

            if request.stream_options and request.stream_options.include_usage:
                created = int(time.time())
                cached_tokens = self._prompt_cache_tokens_count
                logger.debug(f"Stream response with {cached_tokens} cached tokens")

                prompt_tokens_details = None
                if cached_tokens > 0:
                    from ..schema import PromptTokensDetails

                    prompt_tokens_details = PromptTokensDetails(
                        cached_tokens=cached_tokens
                    )

                yield ChatCompletionChunk(
                    id=chat_id,
                    created=created,
                    model=request.model,
                    choices=[
                        ChatCompletionChunkChoice(
                            index=0,
                            delta=ChatMessage(role=Role.ASSISTANT),
                            finish_reason=None,
                            logprobs=None,
                        )
                    ],
                    usage=ChatCompletionUsage(
                        prompt_tokens=result.prompt_tokens + cached_tokens,
                        completion_tokens=result.generation_tokens,
                        total_tokens=result.prompt_tokens
                        + result.generation_tokens
                        + cached_tokens,
                        prompt_tokens_details=prompt_tokens_details,
                    ),
                )

        except Exception as e:
            logger.error(f"Error during stream generation: {str(e)}", exc_info=True)
            raise



from ..text_models import BaseTextModel
from .mlx_model import MLXModel
from .model_types import MlxModelCache, ModelId

# Initialize global cache objects
_model_cache = None
_mlx_model_cache = None


def load_model(model_id: ModelId) -> BaseTextModel:
    """Load the model and return a BaseTextModel instance.

    Args:
        model_id: ModelId object containing model identification parameters

    Returns:
        Initialized BaseTextModel instance
    """
    global _model_cache, _mlx_model_cache

    # Check if a new model needs to be loaded
    model_needs_reload = _model_cache is None or _model_cache.model_id != model_id

    if model_needs_reload:
        # Cache miss, create a new cache object
        _model_cache = MlxModelCache(model_id)

        # Create and cache new MLXModel instance
        _mlx_model_cache = MLXModel(model_cache=_model_cache)

    # Return cached model instance
    return _mlx_model_cache
